import os, sys, re
import xml.dom.minidom
import hashlib
import shutil
import subprocess

def getMD5(filename):
    return hashlib.md5(open(filename,'rb').read()).hexdigest()

def mkdir_recursive(path):
    part1, part2 = os.path.split(path)
    if not os.path.exists(part1):
        mkdir_recursive(part1)
    if not os.path.exists(path):
        os.mkdir(path)

def copyfile_with_create_dir(path_src,path_dest):
    mkdir_recursive(os.path.split(path_dest)[0])
    shutil.copyfile(path_src,path_dest)

def universal_remove(path):
    path = os.path.normpath(path)
    if not os.path.exists(path):
        return
    if os.path.isdir(path):
        shutil.rmtree(path)
    else:
        os.remove(path)

def remove_empty_folders(path):
    path = os.path.normpath(path)
    if not os.path.exists(path):
        print "WARNING (remove_empty_folders): can't allocate path '%s'" % path
        return True
    if os.path.isdir(path):
        is_empty = True
        for item in os.listdir(path):
            if not remove_empty_folders(os.path.join(path, item)):
                is_empty = False
        if is_empty :
            os.rmdir(path)
        return is_empty
    else :
        file_name = os.path.split(path)[1].lower()
        if file_name == '.ds_store' or file_name == 'thumbs.db' :
            universal_remove(path)
            return True
        else:
            return False

def hide_file_or_folder_windows(item_path):
    if (os.path.exists(item_path)):
        exec_cmd_stdout('attrib "{}" +H'.format(os.path.normpath(os.path.abspath(item_path))))
    else:
        print "ERROR: can't allocate item '{}'".format(item_path)
        sys.exit(-1)

def is_power_of_two(N):
    p = 1
    while p<N:
        p = p << 1
    return p==N

def get_power_of_two(N):
    retVal = 1
    while retVal<N:
        retVal = retVal << 1
    return retVal

def make_filelist_stub(path,filter_func,filelist):
    for item in os.listdir(path):
        item_path = os.path.normpath(path + "/" + item)
        #if item_path in ignore_list:
        #    continue
        if os.path.isdir(item_path):
            make_filelist_stub(item_path,filter_func,filelist)
        if filter_func(item_path):
            filelist.append(item_path)

def make_filelist(root_path,filter_func):
    filelist = []
    make_filelist_stub(root_path,filter_func,filelist)
    return filelist

def remove_prefix(s,prefix):
    if(s.startswith(prefix)):
        return s[len(prefix):]
    else:
        return s
def remove_suffix(s,suffix):
    if(s.endswith(suffix)):
        return s[:-len(suffix)]
    else:
        return s

def replace_in_file(file_path, xml_path, output_file_path = ""):
    if not output_file_path :
        output_file_path = file_path
    if not os.path.exists(file_path):
        print "ERROR{replace_in_file}: can't allocate file '%s'" % file_path
        return False
    if not os.path.exists(xml_path):
        print "ERROR{replace_in_file}: can't allocate file '%s'" % xml_path
        return False

    src_file = open(file_path, "r")
    replace_content = src_file.read()
    src_file.close()

    xml_doc = xml.dom.minidom.parse(xml_path)
    items = getXmlNodes_FromNode(xml_doc,"replace_tool/replace_entry")
    for item in items:
        match = str(item.getAttribute("match"))
        replace = str(item.getAttribute("replace"))
        if not match :
            print "ERROR{replace_in_file}: '%s' has entry without match attribute" % xml_path
            return False
        if not replace :
            print "WARNING{replace_in_file}: '%s' has entry without replace attribute. Matches will be just deleted" % xml_path
        replace_content = replace_content.replace(match, replace)

    output_file = open(output_file_path, "w")
    output_file.write(replace_content)
    output_file.close()
    return True

# copy recursively from src to dst. If an item already exists in dest, it will be replaced without any checking
def merge_folder(src, dst, delete_merged_src = False):
    if not os.path.exists(src) or not os.path.isdir(src):
        print "ERROR{merge_folder}: can't allocate source directory '%s'" % src
        return 1
    if not os.path.exists(dst) or not os.path.isdir(dst):
        print "ERROR{merge_folder}: can't allocate destination directory '%s'" % dst
        return 1

    src_items = os.listdir(src)
    for item in src_items :
        src_item_path = os.path.join(src, item)
        dst_item_path = os.path.join(dst, item)
        if os.path.exists(dst_item_path):
            if os.path.isdir(src_item_path) != os.path.isdir(dst_item_path):
                print "ERROR{merge_folder}: can't merge. Both theses pathes shourld be directories:\nsrc:'%s'    \ndst:'%s'" % (src_item_path, dst_item_path)
                return 1
            if os.path.isdir(src_item_path) :
                ret_val = merge_folder(src=src_item_path, dst=dst_item_path, delete_merged_src=delete_merged_src)
                if ret_val :
                    return ret_val
            else :
                os.remove(dst_item_path)
                shutil.copyfile(src_item_path, dst_item_path)
        else :
            if os.path.isdir(src_item_path) :
                mkdir_recursive(dst_item_path)
                ret_val = merge_folder(src=src_item_path, dst=dst_item_path, delete_merged_src=delete_merged_src)
                if ret_val :
                    return ret_val
            else :
                shutil.copyfile(src_item_path, dst_item_path)
        if delete_merged_src :
            universal_remove(src_item_path)
    return 0

def clean_duplicates(clean_path, replacing_path):

    if not os.path.exists(clean_path):
        return (1,["'%s' not a valid path" % clean_path])
    if not os.path.isdir(clean_path):
        return (1,["'%s' not a directory" % clean_path])

    if not os.path.exists(replacing_path):
        return (1,["'%s' not a valid path" % replacing_path])
    if not os.path.isdir(replacing_path):
        return (1,["'%s' not a directory" % replacing_path])

    retval = 0
    output_strings = []

    rep_items = os.listdir(replacing_path)
    for i in rep_items:
        c_path = clean_path + os.path.sep + i
        if os.path.exists(c_path):
            r_path = replacing_path + os.path.sep + i
            if os.path.isdir(r_path):
                call_retval,call_out = clean_duplicates(c_path, r_path)
                if call_retval!=0:
                    retval = 1
                output_strings.extend(call_out)
            else:
                output_strings.append("clean out '%s'" % c_path)
                os.remove(c_path)

    return (retval,output_strings)

def exec_cmd(command):
    #print command
    try:
        output = subprocess.check_output(command, shell=True, stderr=subprocess.STDOUT)
        return 0,output
    except subprocess.CalledProcessError, e:
        return e.returncode,e.output

def exec_cmd_stdout(command):
    #print command
    try:
        subprocess.check_call(command, shell=False, stderr=subprocess.STDOUT)
        return 0
    except subprocess.CalledProcessError, e:
        print(e.output)
        return e.returncode


def mk_dir_recursively(path):
    absfilepath = os.path.abspath(os.path.normpath(path))
    if not os.path.exists(absfilepath):
        path = os.path.split(absfilepath)[0]
        mk_dir_recursively(path)
        os.mkdir(absfilepath)


def mk_dir_for_filepath(filepath):
    absfilepath = os.path.abspath(os.path.normpath(filepath))
    path = os.path.split(absfilepath)[0]
    mk_dir_recursively(path)


def need_refresh(src,dest):
    if not os.path.exists(dest):
        return True
    time_dif = os.path.getmtime(src) - os.path.getmtime(dest)
    if time_dif > 0.0001:
        return True
    if os.path.getsize(src) != os.path.getsize(src):
        return True
    return False

def find_file_in_res_directories(res_folders, sub_path):
    for folder in reversed(res_folders):
        path = os.path.join(folder, sub_path)
        if os.path.exists(path):
            return path
    return ""


def get_video_info(file_path, tools_path):
    # You must have ffmpeg in tools folder
    ffmpeg_path = os.path.join(tools_path, "ffmpeg")
    if os.name != "posix":
        ffmpeg_path += ".exe"
    if not os.path.exists(ffmpeg_path):
        print "Can't allocate ffmpeg at path '%s'"% ffmpeg_path
        return -1
    if not os.path.exists(file_path):
        print "ERROR (get_video_info): Can't locate file '%s'" % (file_path)
        return 0;
    retVal = {}
    command = tools_path + os.path.sep + "ffmpeg -i " + file_path
    retcode, output = exec_cmd(command)

    # resolution
    mask = r', \d+x\d+'
    results = re.findall(mask, output)
    if len(results) != 1:
        print "ERROR (get_video_info): Failed to get info of '%s'.\nCan't find mask '%s' in \n'%s'" % (file_path, mask, output)
        return 0
    result = results[0][2:len(results[0])]
    width = int(result.split('x')[0])
    height = int(result.split('x')[1])
    if width <= 0 or height <= 0:
        print "ERROR:failed to determine resolution of '%s'. Width=%d, height=%d" % (file_path, width, height)
        return 0
    retVal['width'] = width
    retVal['height'] = height
    return retVal

def parse_vector2(str_val):
    list = str_val.split(' ')
    if len(list) != 2:
        print "Failed to parse '%s' into vector2" % str_val
        exit(-1)
    return float(list[0]), float(list[1])

def serialize_vector2(x, y):
    return str("%f"%x).rstrip('0').rstrip('.') + " " + str("%f"%y).rstrip('0').rstrip('.')

def swap_vars(a, b):
    return b, a

def match_regex(str, masks): #takes string and list of regex masks
    for mask in masks :
        if re.match(mask, str, re.I) :
            return True
    return False

def remove_recursivelly_by_mask(path, masks):
    if not os.path.exists(path):
        print "ERROR(remove_recursivelly_by_mask): Can't find path '%s'" % path
        return -1
    if match_regex(path, masks):
        print "Removed:    " + path
        universal_remove(path)
    elif os.path.isdir(path):
        for dir_item in os.listdir(path):
            sub_path = os.path.join(path, dir_item)
            remove_recursivelly_by_mask(sub_path, masks)
    return 0

#=======================================================================================================================
# XML
#=======================================

def parse_xml_file(path, warnings_set = None):
    if not os.path.exists(path):
        error_msg = "ERROR: can't find file '{path}'".format(path=os.path.abspath(os.path.normpath(path)))
        if warnings_set != None:
            warnings_set.add(error_msg)
            return None
        else:
            print error_msg
            sys.exit(-3)

    xml_doc = xml.dom.minidom.parse(path)
    if not xml_doc:
        error_msg = "ERROR: Failed to parse '{path}' as xml".format(path=os.path.abspath(os.path.normpath(path)))
        if warnings_set != None:
            warnings_set.add(error_msg)
            return None
        else:
            print error_msg
            sys.exit(-3)
    return xml_doc

def getXmlNodes_FromNode(node,xml_path):
    subpath_elems = xml_path.split("/")
    nodes = filter(lambda x: x.nodeName == subpath_elems[0], node.childNodes)
    if len(subpath_elems)==1:
        return nodes
    new_subpath = '/'.join(subpath_elems[1:])
    retval = []
    for n in nodes:
        retval = retval + getXmlNodes_FromNode(n,new_subpath)
    return retval

def getXmlNodes_FromFile(filename,xml_path):
    return getXmlNodes_FromNode(xml.dom.minidom.parse(filename),xml_path)

def find_node_by_attr(parent_node, node_path, attr_name, attr_val, warn_on_error_with_msg = None):
    nodes = getXmlNodes_FromNode(parent_node,node_path)
    for node in nodes:
        val = str(node.getAttribute(attr_name))
        if val == attr_val:
            return node
    if warn_on_error_with_msg:
        print warn_on_error_with_msg
    return None

def get_xml_node_children_by_name(xml_node, name):
    retList = []
    for node in xml_node.childNodes:
        if node.nodeType == xml.dom.minidom.Node.ELEMENT_NODE and \
            (name == "*" or node.tagName == name):
            retList.append(node)
    return retList

#=======================================================================================================================

# this methods allows you to build Windows app
def build_using_ms_buld(solution, targets, config, plarform = 'Win32', resource_preprocessor_definitions=''):
    command_template = '"C:\\Program Files (x86)\\Microsoft Visual Studio\\2017\\Community\\MSBuild\\15.0\\Bin\\MSBuild.exe" "{solution}" /m /t:{targets} /p:Configuration="{config}" /p:Platform={plarform}'
    command = command_template.format(solution=solution, targets=targets, config=config, plarform=plarform)
    if resource_preprocessor_definitions:
        command += ' /p:ResourcePreprocessorDefinitions="{}"'.format(resource_preprocessor_definitions)
    print command
    p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    output, err = p.communicate()
    retcode = p.returncode
    if retcode != 0:
        print "MSBuild failed with code {}".format(retcode)
        print "Command: " + command
        print 'Stdout: \n'
        print output
        print 'Stderr: \n'
        print err
        sys.stdout.flush()
        sys.exit(retcode)

#=======================================================================================================================
# Regex ignore
#=======================================
class regex_ignore(object):
    def __init__(self, xml_path):
        xml_doc = parse_xml_file(xml_path)
        self.rule_sections = {}
        if not xml_doc:
            sys.exit(-1)
        root_node = xml_doc.childNodes[0]
        if not root_node:
            print "ERROR: failed to get root node in xml '{path}'".format(path=xml_path)
            sys.exit(-1)
        for section_node in root_node.childNodes:
            section_name = section_node.nodeName
            ignores = []
            exclusions = []
            for rule_node in section_node.childNodes:
                cur_list = None
                if rule_node.nodeName == 'ignore':
                    cur_list = ignores
                elif rule_node.nodeName == 'exclusion':
                    cur_list = exclusions
                if not cur_list == None:
                    mask = str(rule_node.getAttribute("mask"))
                    if not mask :
                        print "ERROR: smart_copy config file has entry without mask"
                        sys.exit(-1)
                    if mask[0] != '^' :
                        mask = '^' + mask
                    if not mask.endswith('$') :
                        mask = mask + '$'
                    cur_list.append(mask)
            self.rule_sections[section_name] = {"ignores":ignores, "exclusions":exclusions}

    def should_ignore(self, check_str, section, verbose = False):
        ignoring = self.check_with_section(check_str, '_common', 'ignores')
        if not ignoring:
            ignoring = self.check_with_section(check_str, section, 'ignores')
        if ignoring:
            ignoring = not self.check_with_section(check_str, '_common', 'exclusions')
        if ignoring:
            ignoring = not self.check_with_section(check_str, section, 'exclusions')

        if ignoring and verbose :
            print "Ignored: {}".format(check_str)
        return ignoring

    def check_with_section(self, check_str, section, sublist_name):
        section_dict = self.rule_sections[section]
        if not section_dict:
            return False
        rules = section_dict[sublist_name]
        for mask in rules :
            if re.match(mask, check_str, re.I) :
                return True
        return False

    """
    re_ignore = regex_ignore("C:/_MyProjects/casual_14/casual14/tools/detect_errors_in_resources_cfg.xml")
    in_list = ['scenes/marmot/extras', '.svn', '..svn', 'bin/.svn', 'bin/tex/back.png', 'paywall', 'scenes/paywall', 'scenes/tex/paywall', 'extras', 'scenes/.extras', 'scenes/extras', 'scenes/tex/extras']
    out_list = filter(lambda x: not re_ignore.should_ignore(x,'marmot', True), in_list)
    print "\n".join(out_list)
    """















